# Dynamic Buckets Plugin

**Supported Sisense Version**: V5.7.7  
**Plugin Version**: 2.2  
**Last updated**: 15 October 2018

## Description

This plugin helps when you need to group x-Axis into buckets according to first value. Bucket size (y-Axis) is a sum of the second value

## Installation

1. Download the plugin
2. Extract the `.zip` folder to `C:\Program Files\Sisense\PrismWeb\Plugins` (For Sisense v7.2+ extract to `C:\Program Files\Sisense\app\plugins`)
    - If the folder does not exist, please create it before extracting the `.zip` file.
3. Add configuration script (read more below) at "Edit script" window of widget
4. Refresh your dashboard

## Implementation

1. Add the following widget script for the widget you would like the plugin functionality to apply on:

    - Example 1: Manually define ranges

    ```javascript
    // Define the plugin options
    var options = {
        type: "manual",            // (required) Can be 'auto' or 'manual'
        separator: " - ",          // (optional) If the ranges aren't named, create one using the min & max values and this separator
        clickBucketOption: 1,      // (optional) Can be 1 (default) or 2. 1- sets column categories as filters, 2- exports column categories to CSV file.
        ranges: [                  // Manually define the buckets you want to use
            // This will group any values less or equal to 1
            {
                name: "Bad",       // Label for the bucket
                color: "red",      // Color can be text or a hex code
                min: null,         // The lower limit for the bucket, put null for no lower limit
                max: 1             // The upper limit for the bucket, put null for no upper limit
            },
            // This will group any values between 1 - 10
            {
                name: "OK",
                color: "yellow",
                min: 1,
                max: 10
            },
            // This will group any values greater than 10
            {
                name: "Good",
                color: "green",
                min: 10,
                max: null
            }
        ]
    };

    // Run the function to create buckets
    prism.dynamicBuckets(widget, options);
    ```

    - Example 2: Automatically define ranges

    ```javascript
    // Define the plugin options
    var options = {
        type: "auto",               // (required) Can be 'auto' or 'manual'
        numberOfBuckets: 2,         // (required) How many buckets should we calculate
        separator: " - ",           // (optional) If the ranges aren't named, create one using the min & max values and this separator
        clickBucketOption: 1,       // (optional) Can be 1 (default) or 2. 1- sets column categories as filters, 2- exports column categories to CSV file
        normalizeFloats: false      // (optional) If true, forces buckets with floating point ranges into reliable values rounded to one significant figure
    };

    // Run the function to create buckets
    prism.dynamicBuckets(widget, options);
    ```

## Limitations

- Multiple selections are not supported for this plugin
