prism.run([
    'plugin-dynamicBuckets.services.bindOnce',
    function ($bindOnce) {

        prism.dynamicBuckets = function (widget, options) {
            widget.dynamicBucketsOptions = options;
        };

        prism.on('dashboardloaded', function (se, args) {

            // init Dynamic Bucket for widgets were initialized via widget script (need for PhantomJS)
            args.dashboard.widgets.toArray().forEach(function (w) {
                if(w.dynamicBucketsOptions){
                   initDynamicBucket(w);
                }
            });

            args.dashboard.on("widgetadded", function (e, args) {
                if(args.reason === "duplicate" && args.widget.dynamicBucketsOptions){
                    // do not show widget to hide view changing from current to dynamic buckets
                    $("[widgetid='" + args.widget.oid + "']").css({"opacity": 0});
                    initDynamicBucket(args.widget);
                    // must be refreshed to change widget view for widget new model of dynamic buckets
                    args.widget.refresh();
                }
            });

            // for entering widget edit mode and exiting back to dashboard view from it
            args.dashboard.on('widgetinitialized', function(se, args) {
                if(args.widget.dynamicBucketsOptions){
                    initDynamicBucket(args.widget);
                }
            })
        });

        // Allows region mouse "select" and "unselect" filtering.
        $bindOnce(prism, "beforemenu", beforeMenuHandler);

        function initDynamicBucket(widget) {

            //	Setting data limitation to 2000000, rendering performance might suffer from client side overload
            $bindOnce(widget, "buildquery", buildQueryHandler);

            //	Transform the result set into buckets
            $bindOnce(widget, "processresult", processResultHandler);
        }

        //	Setting data limitation to 2000000, rendering performance might suffer from client side overload
        function buildQueryHandler(widget, ev) {

            //	Don't do anything, if this is a pivot table
            if (widget.type === 'pivot') {
                return;
            }
            //	Set the row count limit
            ev.query.count = 2000000;
        }

        //	Transform the result set into buckets
        function processResultHandler(widget, ev) {
            // show temporally hidden widget
            $("[widgetid='" + widget.oid + "']").css({"opacity": 1});

            var options = widget.dynamicBucketsOptions;

            //	Check if 2 values are defined
            if (!ev.result.series || ev.result.series.length < 2) {
                return;
            }

            //	Create bucket groups
            var buckets = [];
            ev.result.xAxis.categories = [];

            //	Check for bucket type
            var type = options.type;

            //	Add formatting to the options object
            options.formatting = widget.metadata.panels[1].items[0].format.mask;

            //	Create buckets differently for each type
            switch (type) {
                case "manual":
                    buckets = calculateBucketsManually(ev, options);
                    break;
                case "auto":
                    buckets = calculateBucketsAuto(ev, options);
                    break;
                default:
                    break;
            }


            //	Set bucket values in the chart object
            createBuckets(ev, buckets);

        }

        // Allows region mouse "select" and "unselect" filtering.
        function beforeMenuHandler(ev, args) {

            // do not select if bucket is empty (removed menu item)
            var selectedMembersCount = $$get(args, 'settings.jaql.0.jaql.filter.members.length');
            if (selectedMembersCount === 0) {
                var selectItem = _.find(args.settings.items, function(curItem){
                    return curItem.mid === 'select' || curItem.mid === 'unselect';
                });
                selectItem.disabled = 1;
                return;
            }

            // Update menu if "datapoint" menu opened and widget in bucket mode
            if (args.settings.name === "datapoint" && args.settings.widget.dynamicBucketsOptions && prism.$ngscope.appstate !== 'widget') {

                var selectOperation = "select";
                var unSelectOperation = "unselect";
                // FIX [PSE-1450] Handle sporadic missing 'pointDetails' object
                var series = !defined(args.pointDetails) ? args.settings.widget.queryResult.series[0] : args.pointDetails.pointScope.series;
                var filters = args.settings.jaql;
                var jaql = filters[0].jaql;

                // check if dimension is datetime
                var isDate = jaql.datatype === "datetime";

                // Collect "selected data" for all buckets in series
                var selectionData = series.data.map(function (item) {
                    return item.selectionData;
                });

                // Get "selected data" only for selected buckets
                var selectionDataFiltered = selectionData
                    .map(function(dateStr) {
                        return isDate && dateStr ? dateStr.map(function (dateItem) {
                            return new Date(dateItem).getTime();
                        }) : dateStr;
                    }).filter(function (group) {
                        return jaql.filter.members.some(function (filterValue) {
                            return group.indexOf(isDate ? filterValue.getTime() : filterValue) !== -1;
                        });
                    });

                // Collect all "selected data" for selected buckets
                jaql.filter.members = selectionDataFiltered.reduce(function (acc, item) {
                    // return acc.concat(item);
                    return acc.concat(!isDate ? item : item.map(function(elem){
                        return new Date(elem);
                    }));
                }, []);

                // Find "select" or "unselect" menu item
                var actionMenuItem = _.find(args.settings.items, function (item) {
                    return item.mid === selectOperation || item.mid === unSelectOperation;
                });

                var captionMenuItem = args.settings.items[0].type === "header" ? args.settings.items[0] : null;

                // Change header menu caption
                if (captionMenuItem) {

                    if (jaql.filter.members.length && jaql.filter.members.length <= 2) {
                        // If one or two values in filter, then show their values as caption
                        captionMenuItem.caption = jaql.filter.members.join(", ");
                    }
                    else {
                        // Set dim title as a caption
                        captionMenuItem.caption = jaql.title;
                    }
                }

                jaql.filter.multiSelection = jaql.filter.members.length > 1;
                jaql.collapsed = true;

                // Break if no required action or if widget not affect dashboard filters
                if (!actionMenuItem || !args.settings.widget.options.selector) return;

                // Set "select" execution handler
                if (actionMenuItem.mid === selectOperation) {

                    actionMenuItem.execute = function () {
                        performFilterSelect(prism.activeDashboard, filters);
                    };
                }

                // Set "unselect" execution handler
                if (actionMenuItem.mid === unSelectOperation) {

                    actionMenuItem.execute = function () {
                        performFilterUnSelect(prism.activeDashboard, filters);
                    };
                }

                // Clear default behaviour
                actionMenuItem.mid = "none";

            }
        }

        // Function to set/unset a dashboard filter, when a user clicks on a bucket
        function setFilter(widget) {

            var dashboard = prism.activeDashboard;
            var filters = [];

            //	Get the filter jaql
            var jaql = this.selectionDim;

            // Break if not in dashboard or widget not effects on dashboard filters
            if (prism.$ngscope.appstate && prism.$ngscope.appstate !== "dashboard" || !widget.options.selector) return;

            //	Create a filter object
            jaql.filter = {
                explicit: true,
                multiSelection: true,
                members: this.selectionData
            };

            //	Modify the jaql to include the filter selection(s)
            jaql.collapsed = true;
            delete jaql.in;
            delete jaql.merged;

            // Create prepared filter object
            filters.push({jaql: jaql});

            // Check if currently clicked bucket is already selected in dashboard filters
            var isSameFilterSelected = dashboard.filters.$$items.some(function (filter) {

                var currentJaql = filter.jaql;

                // If cascading filters selected, then search for required dim
                if (filter.levels) {
                    currentJaql = _.find(filter.levels, function (item) {
                        return item.dim === jaql.dim;
                    });
                }

                if (currentJaql && currentJaql.dim === jaql.dim && currentJaql.filter.members) {
                    var membersNew = currentJaql.filter.members.slice();
                    var membersOld = jaql.filter.members.slice();

                    membersNew.sort();
                    membersOld.sort();

                    // compare old and new filter items
                    return membersNew.join("") === membersOld.join("");
                }
                else {
                    return false;
                }

            });

            if (isSameFilterSelected) {
                // if same bucket already selected then remove it from the dashboard filters
                performFilterUnSelect(dashboard, filters);
            }
            else {
                // otherwise select current bucket for the dashboard filters
                performFilterSelect(dashboard, filters);
            }
        }

        function exportToCSV(item, a) {
            if (this.selectionData) {
                var csv = this.selectionDim.title + '\n';
                csv += this.selectionData.join('\n');
                var csvData = new Blob([csv], {type: 'text/csv'});
                var csvUrl = URL.createObjectURL(csvData);

                var fileName = 'widget data';
                var link = document.createElement("a");
                link.href = csvUrl;
                link.style = "visibility:hidden";
                link.download = fileName + ".csv";
                document.body.appendChild(link);
                link.click();
                document.body.removeChild(link);
            }
        }

        //	Create buckets from a manually entered range
        function calculateBucketsManually(ev, options) {

            //	Start with an empty array for the buckets
            var buckets = [];

            //	Start with an empty array for the categories
            var categories = [];

            //	Get the range specified by the user
            var ranges = options.ranges;

            //	Loop through each data series
            for (var j = 0; j < ev.result.series.length - 1; j++) {

                //	Create an array for this data series
                buckets[j] = [];

                //	Make sure a separator is defined
                var separator = options.separator ? options.separator : "-";

                //	Create a new range for this data series
                var myRange = $.extend(true, [], ranges);

                //	Loop through each value in this series
                for (var i = 0; i < myRange.length; i++) {

                    //	Add some extra properties to the range
                    myRange[i].members = [];
                    myRange[i].count = 0;
                    myRange[i].selectedFlags = [];

                    //	Save the current range to the buckets list
                    buckets[j].push(myRange[i]);

                    //	Figure out the label to use
                    var label = myRange[i].name ? myRange[i].name : (min == null ? '' : min) + separator + (max == null ? '' : max);

                    //	Only need to add the axis labels once
                    if (j == 0) {
                        //	Show the label on the x-axis
                        ev.result.xAxis.categories.push(label);
                    }
                }
            }

            //	Return the finished buckets object
            return buckets;
        }

        function normalizeXAxisRanges(xAxis) {
            xAxis.categories = xAxis.categories.map(function(c) {
                var catArray = c.split(' ');
                var separator = catArray.splice(1, 1);
                return catArray[0].substr(0, 3) + ' ' + separator + ' ' + catArray[1].substr(0, 3);
            });
        }

        function floatingPointRound(float) {
            return Math.round(float * 10) / 10;
        }

        //  Normalize floating point math on coefficients and float-dependent buckets
        function normalizeFloat(series) {
            return series.map(function(s) {
                return s.map(function(bucket) {
                    if (bucket.min < 1.1) bucket.min = floatingPointRound(bucket.min);
                    if (bucket.max < 1.1) bucket.max = floatingPointRound(bucket.max);
                    return bucket;
                });
            });
        }

        //	Calculate buckets based on the values
        function calculateBucketsAuto(ev, options) {

            //	The number of buckets to create
            var number_of_buckets = options.numberOfBuckets;

            //	The separator to use
            var separator = options.separator ? options.separator : " - ";

            //	Get the data series
            var data = ev.result.series[0].data;

            //	Start with an empty list for the buckets
            var buckets = [];

            //	Pick a default minimum and maximum
            var max_value_y_axis = Math.ceil(+(data[data.length - 1].y));
            var min_value_y_axis = Math.floor(+(data[0].y));

            //	Loop through each data point
            data.forEach(function (e) {

                //	Check for the max value (rounded up)
                if (e.y > max_value_y_axis) {
                    max_value_y_axis = Math.ceil(+(e.y));
                }

                //	Check for the min value (rounded down)
                if (e.y < min_value_y_axis) {
                    min_value_y_axis = Math.floor(+(e.y));
                }
            });

            //	Calculate the intervals to make x number of buckets
            var range = (max_value_y_axis - min_value_y_axis);
            var unroundedTickSize = range / (number_of_buckets);
            var x = Math.ceil(Math.log10(unroundedTickSize) - 1);
            var pow10x = Math.pow(10, x);
            // var bucket_interval = Math.ceil((max_value_y_axis - min_value_y_axis) / number_of_buckets);
            var bucket_interval = Math.ceil(unroundedTickSize / pow10x) * pow10x;

            if ((bucket_interval * (number_of_buckets - 1)) > max_value_y_axis) {
                bucket_interval = bucket_interval * (number_of_buckets - 1) / number_of_buckets;
            }

            //	Loop through each data series
            for (var i = 0; i < ev.result.series.length - 1; i++) {

                //	Add an empty array for this series
                buckets[i] = [];

                //	Loop through data points in this series
                for (var j = 0; j < number_of_buckets; j++) {

                    //	Calculate the min and max values
                    var min = min_value_y_axis + (bucket_interval * j);
                    var max = min_value_y_axis + bucket_interval * (j + 1);

                    //	Add a bucket object for this point
                    buckets[i].push({min: min, max: max, count: 0, members: [], selectedFlags: []});

                    //	Add the labels if necessary
                    if (!i) {

                        //	Create a label using the defined formatting
                        var label = formatter(min, options.formatting) + separator + formatter(max, options.formatting);
                        //ev.result.xAxis.categories.push(min + separator + max);
                        ev.result.xAxis.categories.push(label);
                    }
                }
            }

            //	Return the buckets object
            if (options.normalizeFloats) {

                // Normalize floating point numbers if specified in options
                normalizeXAxisRanges(ev.result.xAxis);
                return normalizeFloat(buckets);
            }

            return buckets;
        }

        function formatter(originalValue, format) {

            //	Grab the original value
            var value = originalValue;

            //	Init the label
            var label = "";

            //	Use variables for rounding thresholds
            var trillion = 1000000000000;
            var billion = 1000000000;
            var million = 1000000;
            var thousand = 1000;

            //	Handle decimal places
            var decimalPlaces = format.decimals == 'auto' ? 1 : format.decimals;
            var decimalPadding = Math.pow(10, decimalPlaces);

            //	Create variables for prefix and postfix
            var prefix = '';
            var postfix = '';

            //	Apply rounding, if necessary
            if (format.abbreviations) {
                if (value > trillion) {
                    //	Apply rounding up to trillions
                    if (format.abbreviations.t) {
                        value = Math.round(decimalPadding * value / trillion) / decimalPadding;
                        postfix = 'T';
                    } else if (format.abbreviations.b) {
                        value = Math.round(decimalPadding * value / billion) / decimalPadding;
                        postfix = 'B';
                    } else if (format.abbreviations.m) {
                        value = Math.round(decimalPadding * value / million) / decimalPadding;
                        postfix = 'M';
                    } else if (format.abbreviations.k) {
                        value = Math.round(decimalPadding * value / thousand) / decimalPadding;
                        postfix = 'K';
                    }
                } else if (value > billion) {
                    //	Apply rounding up to billions
                    if (format.abbreviations.b) {
                        value = Math.round(decimalPadding * value / billion) / decimalPadding;
                        postfix = 'B';
                    } else if (format.abbreviations.m) {
                        value = Math.round(decimalPadding * value / million) / decimalPadding;
                        postfix = 'M';
                    } else if (format.abbreviations.k) {
                        value = Math.round(decimalPadding * value / thousand) / decimalPadding;
                        postfix = 'K';
                    }
                } else if (value > million) {
                    //	Apply rounding up to millions
                    if (format.abbreviations.m) {
                        value = Math.round(decimalPadding * value / million) / decimalPadding;
                        postfix = 'M';
                    } else if (format.abbreviations.k) {
                        value = Math.round(decimalPadding * value / thousand) / decimalPadding;
                        postfix = 'K';
                    }
                } else if (value > thousand) {
                    //	Apply rounding up to thousands
                    if (format.abbreviations.k) {
                        value = Math.round(decimalPadding * value / thousand) / decimalPadding;
                        postfix = 'K';
                    }
                }
            }

            //	Check for currency
            if (format.currency) {
                if (format.currency.position == "pre") {
                    prefix = format.currency.symbol;
                } else {
                    postfix = postfix + ' ' + format.currency.symbol;
                }
            }

            //	Check for percent
            if (format.percent) {
                value = Math.round(value * 100 * decimalPadding) / decimalPadding;
                postfix = '%';
            }

            //	Set the # of decimal places
            if (format.decimals != "auto") {
                value = value.toFixed(decimalPlaces);
            }

            //	Create a label for the value
            value = prefix + value + postfix;

            //	Return the formatted value
            return value;
        }

        //	Group the members into buckets, update the chart
        function createBuckets(ev, buckets) {

            var options = ev.widget.dynamicBucketsOptions;

            var clickOptionHandler = function () {
                setFilter.call(this, ev.widget);
            };

            // If export to csv selected
            if (options.clickBucketOption === 2) {
                clickOptionHandler = exportToCSV;
            }

            //	Save the dim for filters later
            var filterDim = ev.rawResult.metadata[0].jaql;

            //	Loop through each series and add values to each bucket
            for (var series_number = 0; series_number < ev.result.series.length - 1; series_number++) {

                //	Loop through each value in this series
                for (var i = 0; i < ev.result.series[series_number + 1].data.length; i++) {

                    //	Get the data point
                    var y_axis_val = ev.result.series[0].data[i].y;

                    //	Loop through each bucket to see where to put this data point
                    for (var j = 0; j < buckets[series_number].length; j++) {

                        //	Check if value is less than the max, and there is no min for this bucket
                        var check1 = (buckets[series_number][j].min == null && y_axis_val < buckets[series_number][j].max);

                        //	Check if value is greater than the min, and there is no max for this bucket
                        var check2 = (buckets[series_number][j].max == null && y_axis_val >= buckets[series_number][j].min);

                        //	Check if value is between min & max for this bucket
                        var check3 = ((buckets[series_number][j].min != null && y_axis_val >= buckets[series_number][j].min) && (buckets[series_number][j].max != null && y_axis_val < buckets[series_number][j].max));

                        //	Did any of the checks pass?
                        if (check1 || check2 || check3) {

                            //	Increment the count for this bucket
                            buckets[series_number][j].count += ev.result.series[series_number + 1].data[i].y;

                            //	Save the member, so we can filter if a user clicks on this group
                            var selectionDataValue = ev.result.series[series_number + 1].data[i].selectionData[0];
                            //  FIX [PSE-1411] If no min lower boundary and y_axis_val is 0, ensure pushed value is '0' and not 'null'
                            var valueToPush = (y_axis_val === 0 && selectionDataValue === null) ? 0 : selectionDataValue;
                            buckets[series_number][j].members.push(valueToPush);

                            //	Save the "highlight" value for each member
                            buckets[series_number][j].selectedFlags.push(ev.result.series[series_number + 1].data[i].selected);
                        }
                    }
                }
            }

            //	Loop through each series in the chart and update y-axis values for each bucket
            for (var series_number = 1; series_number < ev.result.series.length; series_number++) {

                //	Remove the old series' data, and start with an empty array
                delete ev.result.series[series_number].data;
                ev.result.series[series_number].data = [];

                //	Loop through each bucket, and add them as new data points
                for (var i = 0; i < buckets[series_number - 1].length; i++) {

                    // Data point active if any member is active (selected === false) in case of "highlight" filtering
                    var isDataPointActive = buckets[series_number - 1][i].selectedFlags.some(function (selected) {
                        return selected === false;
                    }) || !buckets[series_number - 1][i].selectedFlags.length;

                    //	Create a new data point for this bucket
                    var dataPoint = {
                        y: buckets[series_number - 1][i].count,					//	The new value displayed will be the count
                        color: buckets[series_number - 1][i].color,				//	Save the color for this bucket
                        selectionData: buckets[series_number - 1][i].members,	//	Save the list of members for this bucket
                        selectionDim: filterDim,								//	Save the filter JAQL
                        events: {
                            //click: setFilter									//	Override the click event for the chart, so that a filter can be added
                            click: clickOptionHandler
                        },
                        selected: !isDataPointActive							//  Set value for "highlight" filtering ("selected === false" for active column)
                    };

                    //	Add the new data point
                    ev.result.series[series_number].data.push(dataPoint);
                }
            }

            //	Get rid of the very first data series, as it was only used included to calculate the groups
            ev.result.series.splice(0, 1);
        }

        // Add/update the filter for the dashboard
        function performFilterSelect(dashboard, filters) {

            var options = {
                refresh: true,
                save: true,
                unionIfSameDimensionAndSameType: false,
                reason: 'widgetSelection'
            };

            dashboard.filters.update(filters, options);
        }

        // Remove/update the filter for the dashboard
        function performFilterUnSelect(dashboard, filters) {

            var options = {
                unUnionIfSameDimensionAndSameType: true,
                reason: 'widgetSelection'
            };

            angular.forEach(filters, function (f) {
                dashboard.filters.remove(f, options);
            });
            dashboard.refresh('dashboardFiltersChanged', filters);
            dashboard.$dashboard.updateDashboard(dashboard, "filters");
        }

    }
]);